from typing import List, Optional

from werkzeug.utils import cached_property

from kama_sdk.core.core.types import KoD
from kama_sdk.model.base.model import Model
from kama_sdk.model.operation.stage import Stage


class Operation(Model):

  @cached_property
  def synopsis(self) -> str:
    return self.get_prop(SYNOPSIS_KEY) or self.title

  @cached_property
  def stages(self) -> List[Stage]:
    """
    Loads the Stages associated with the Operation.
    :return: list of Stage instances.
    """
    return self.inflate_children(Stage, prop=STAGES_KEY)

  def stage(self, stage_id: str) -> Stage:
    """
    Finds the Stage by key and inflates (instantiates) into a Stage instance.
    :param stage_id: identifier for desired Stage.
    :return: Stage instance.
    """
    matcher = lambda stage: stage.id() == stage_id
    return next(filter(matcher, self.stages), None)

  def is_system(self) -> bool:
    return self.id() in ['installation', 'uninstall']

  @cached_property
  def preflight_action_kod(self) -> Optional[KoD]:
    return self.config.get('preflight_action')
    # todo likely wrong should be just get_prop() in case value is supplied


STAGES_KEY = 'stages'
SYNOPSIS_KEY = 'synopsis'
PREFLIGHT_PREDS_KEY = 'preflight_predicates'



from typing import Optional, Dict, List, Union, Any

from typing_extensions import TypedDict


class InjectionsDesc(TypedDict):
  standard: Dict
  inline: Dict


class TemplateArgs(TypedDict, total=False):
  flat_inlines: Dict
  values: Dict


class InputOption(TypedDict, total=False):
  id: str
  title: str


class ActionStatusDict(TypedDict, total=False):
  id: Optional[str]
  title: str
  info: Optional[str]
  status: str
  sub_items: List['ActionStatusDict']
  data: Dict
  error: Dict
  error_id: str
  logs: List[str]


class PromMatrixEntry(TypedDict):
  metric: Dict
  values: List[Union[str, int]]


class PromVectorEntry(TypedDict):
  metric: Dict
  value: List[Union[str, int]]


PromMatrix = List[PromMatrixEntry]


PromVector = List[PromVectorEntry]


class PromData(TypedDict):
  resultType: Dict
  result: Union[PromMatrix, PromVector]


class PromSeriesDataPoint(TypedDict):
  timestamp: str
  value: float


class PortForwardSpec(TypedDict):
  namespace: str
  pod_name: str
  pod_port: int


class EndpointDict(TypedDict):
  url: Optional[str]
  svc_type: Optional[str]
  port_forward_spec: Optional[PortForwardSpec]


class UpdateDict(TypedDict):
  id: str
  version: str
  ktea_type: Optional[str]
  ktea_uri: Optional[str]
  note: str


class JobStatusPart(TypedDict):
  name: str
  status: str
  pct: Optional[int]


class JobStatus(TypedDict):
  parts: List[JobStatusPart]
  logs: List[str]


class PredEval(TypedDict, total=False):
  predicate_id: str
  name: str
  met: bool
  reason: Optional[str]
  tone: str


class ExitStatuses(TypedDict, total=False):
  positive: List[PredEval]
  negative: List[PredEval]


class CommitOutcome(TypedDict, total=False):
  status: str
  reason: Optional[str]
  logs: List[str]


class K8sResSig(TypedDict):
  kind: str
  name: str


class K8sResMeta(TypedDict):
  namespace: str
  name: str


class K8sResDict(TypedDict):
  kind: str
  metadata: K8sResMeta


class KteaDict(TypedDict, total=False):
  type: str
  uri: str
  args: Optional[str]
  version: str
  weak_merge: Optional[Dict]
  strong_merge: Optional[Dict]


class KamaDict(TypedDict, total=False):
  type: str
  uri: str
  version: str


class ConfigBackup(TypedDict, total=False):
  event_id: str
  timestamp: str
  data: Dict


class ActionOutcome(TypedDict):
  cls_name: str
  id: str
  charge: str
  data: Dict


class StepActionKwargs(TypedDict):
  inline_assigns: Dict
  chart_assigns: Dict
  state_assigns: Dict


class KAO(TypedDict):
  api_group: Optional[str]
  kind: str
  name: str
  verb: Optional[str]
  error: Optional[str]


class ErrCapture(TypedDict, total=False):
  vid: str
  event_vid: Optional[str]
  type: str
  name: str
  reason: str
  fatal: bool
  logs: List[str]
  is_original: bool
  extras: Dict[str, Any]
  synced: bool
  occurred_at: str


class EventCapture(TypedDict):
  vid: str
  parent_vid: Optional[str]
  type: str
  name: str
  status: str
  logs: List[str]
  occurred_at: str


class SnapshotMetric(TypedDict):
  type: str
  data: Union[Dict, List]


class NamespacedSnapshot(TypedDict):
  namespace: str
  resources: List
  metrics: List[SnapshotMetric]


class Snapshot(TypedDict):
  namespaced_snapshots: List[NamespacedSnapshot]


class SimpleSeriesSummary(TypedDict):
  series: List[PromSeriesDataPoint]
  humanized_series: List[Dict]
  direction: str
  good_direction: str
  summary_value: str


class ConcernAttrMeta(TypedDict, total=False):
  label: str
  title: str


ConcernRowAttrs = List[ConcernAttrMeta]


class PodStatusSummary(TypedDict):
  ternary_status: str
  pod_name: str
  phase: str

# class PodStatusesSummary(TypedDict)


KAOs = List[KAO]
KoD = Union[str, dict]
KDLoS = Union[str, dict]


class Reconstructor(TypedDict):
  adapter_ref: Union[str, Dict]
  seed: Dict



from typing import List, Dict, Union

from typing_extensions import TypedDict
from werkzeug.utils import cached_property

from kama_sdk.core.core import consts
from kama_sdk.core.core.types import ActionStatusDict, KoD, ErrCapture
from kama_sdk.model.action.base.action import Action
from kama_sdk.model.action.base.action_errors import FatalActionError


class SubActionRule(TypedDict, total=False):
  action: KoD
  patch: Union[str, List[str]]
  _pass: str


class MultiAction(Action):

  SUB_ACTIONS_KEY = 'sub_actions'

  @cached_property
  def sub_actions(self) -> List[Action]:
    return self.inflate_children(
      Action,
      prop=self.SUB_ACTIONS_KEY
    )

  def circuit_breakers(self) -> List[Dict]:
    return self.get_prop('circuit_breakers', [])

  def final_sub_actions(self) -> List[Action]:
    return self.sub_actions

  def perform(self):
    results = {}
    for index, action in enumerate(self.final_sub_actions()):
      ret = action.run()
      if issubclass(ret.__class__, Dict):
        results = {**results, **ret}
        self.patch(ret)
      term = self.on_sub_action_finished(action, index, ret)
      if term == consts.pos:
        return results
      elif term == consts.neg:
        raise FatalActionError(ErrCapture(
          type='early_exit',
          reason='Early exit'
        ))
    return results

  def on_sub_action_finished(self, action: Action, index, ret):
    matcher = lambda cb: cb_matches_action(cb, action, index)
    triggered_breakers = list(filter(matcher, self.circuit_breakers()))
    for breaker in triggered_breakers:
      exit_status = self.resolve_prop_value(breaker.get('exit'))
      if exit_status:
        return exit_status
    return None

  def serialize_progress(self) -> ActionStatusDict:
    sub_actions = self.final_sub_actions()
    sub_progress = [a.serialize_progress() for a in sub_actions]
    own_progress = super().serialize_progress()
    return {
      **own_progress,
      **{'sub_items': sub_progress}
    }


def gen_run_kwargs(action: Action, results, prev) -> Dict:
  keys = action.expected_run_args
  pool = {**prev, **results}.items()
  result = {k: v for k, v in pool if k in keys}
  missing = list(keys - result.keys())
  if len(missing) > 0:
    print(f"[action:{action.id}] invoke with missing kw {missing}")
  return result


def cb_matches_action(breaker_desc: Dict, action: Action, index: int) -> bool:
  after_desc = breaker_desc.get('after')
  if after_desc:
    if after_desc == action.id():
      return True
    elif type(after_desc) == dict:
      return after_desc.get('id') == action.id() or \
             int(after_desc.get('index', -1)) == index
  else:
    return True



from typing import List, Dict, TypeVar

from k8kat.res.base.kat_res import KatRes
from k8kat.utils.main.api_defs_man import api_defs_man
from werkzeug.utils import cached_property

from kama_sdk.core.core import utils
from kama_sdk.core.core.config_man import config_man
from kama_sdk.core.core.types import K8sResDict
from kama_sdk.core.core.utils import deep_get2
from kama_sdk.model.base.model import Model

T = TypeVar('T', bound='ResourceSelector')


class ResourceSelector(Model):

  @classmethod
  def inflate_with_literal(cls, string: str, **kwargs) -> T:
    parts = string.split(':')
    return cls.inflate_with_config({
      RES_KIND_KEY: parts[len(parts) - 2],
      RES_NAME_KEY: parts[len(parts) - 1]
    }, **kwargs)

  @cached_property
  def res_kind(self) -> str:
    return self.resolve_prop(RES_KIND_KEY, missing='raise')

  @cached_property
  def res_name(self) -> str:
    return self.resolve_prop(RES_NAME_KEY)

  def res_namespace(self) -> str:
    return self.get_prop(RES_NS_KEY, config_man.ns())

  @cached_property
  def label_selector(self) -> Dict:
    return self.resolve_prop(LABEL_SEL_KEY, depth=1) or {}

  @cached_property
  def not_label_selector(self) -> Dict:
    return self.resolve_prop(NOT_LABEL_SEL_KEY, depth=1) or {}

  @cached_property
  def field_selector(self) -> Dict:
    return self.resolve_prop(FIELD_SEL_KEY, depth=1) or {}

  @cached_property
  def kat_prop_selector(self):
    return self.resolve_prop(PROP_SEL_KEY, depth=1) or {}

  def query_cluster(self) -> List[KatRes]:
    kat_class: KatRes = KatRes.class_for(self.res_kind)
    if kat_class:
      query_params = self.build_k8kat_query()
      if not kat_class.is_namespaced():
        del query_params['ns']
      return kat_class.list(**query_params)
    else:
      print(f"[kama_sdk::resourceselector] DANGER no kat for {self.res_kind}")
      return []

  def selects_kat_res(self, kat_res: KatRes) -> bool:
    return self.selects_res(utils.kres2dict(kat_res))

  def selects_res(self, res: K8sResDict) -> bool:
    res = fluff_resdict(res)

    kinds1 = api_defs_man.kind2plurname(self.res_kind)
    kinds2 = api_defs_man.kind2plurname(res['kind'])

    if kinds1 == kinds2 or self.res_kind == '*':
      query_dict = self.build_k8kat_query()
      res_labels = (res.get('metadata') or {}).get('labels') or {}
      labels_match = query_dict['labels'].items() <= res_labels.items()
      fields_match = keyed_compare(query_dict['fields'], res)
      return labels_match and fields_match
      pass
    else:
      return False

  def build_k8kat_query(self) -> Dict:
    field_selector = self.field_selector

    if self.res_name and self.res_name != '*':
      field_selector = {
        'metadata.name': self.res_name,
        **(field_selector or {}),
      }

    return dict(
      ns=self.res_namespace(),
      labels=self.label_selector,
      not_labels=self.not_label_selector,
      fields=field_selector
    )


def fluff_resdict(resdict: Dict) -> Dict:
  if 'metadata' not in resdict.keys():
    if 'name' in resdict.keys():
      return dict(
        **resdict,
        metadata=dict(name=resdict['name'])
      )
  return resdict


def keyed_compare(keyed_q_dict: Dict, against_dict: Dict) -> bool:
  for deep_key, check_val in keyed_q_dict.items():
    actual = deep_get2(against_dict, deep_key)
    if not actual == check_val:
      return False
  return True


RES_KIND_KEY = 'res_kind'
RES_NAME_KEY = 'name'
RES_NS_KEY = 'namespace'
LABEL_SEL_KEY = 'label_selector'
NOT_LABEL_SEL_KEY = 'not_label_selector'
FIELD_SEL_KEY = 'field_selector'
PROP_SEL_KEY = 'prop_selector'



from typing import Dict

from kama_sdk.core.core.config_man import config_man
from kama_sdk.model.operation.field import Field
from kama_sdk.model.input import input_serializer
from kama_sdk.model.operation.operation_state import OperationState
from kama_sdk.model.operation.step import Step


def ser_embedded_field(field: Field) -> Dict:
  current_or_default = field.variable.current_or_default_value()

  return dict(
    id=field.id(),
    title=field.title,
    info=field.info,
    is_inline=field.is_inline_chart_var(),
    default=current_or_default,
    decorated_value=field.variable.compute_decoration(),
    **input_serializer.in_variable(field.variable.input_model),
  )


def ser_refreshed(step: Step, values: Dict, state: OperationState) -> Dict:
  """
  Standard serializer for a step.
  :param step: Step class instance.
  :param values: current user input
  :param state: current operation state
  :return: serialized Step in dict form.
  """
  parent_id = step.parent.id() if step.parent else None
  config_man.user_vars()
  visible_fields = step.visible_fields(values, state)
  return dict(
    id=step.id(),
    title=step.title,
    info=step.info,
    flags=[],
    stage_id=parent_id,
    fields=list(map(ser_embedded_field, visible_fields))
  )



from kama_sdk.model.input import input_serializer
from kama_sdk.model.variable.manifest_variable import ManifestVariable


def standard(cv: ManifestVariable, **kwargs):
  return dict(
    id=cv.id(),
    title=cv.title,
    mode=cv.mode,
    description=cv.info,
    value=cv.current_value(**kwargs),
    tags=cv.tags,
    sources=cv.sources(),
    is_valid=cv.is_currently_valid()
  )


def full(cv: ManifestVariable):
  """
  Extended serializer for the ChartVariable instance, which also includes includes
  details about the associated field.
  :param cv: ChartVariable class instance.
  :return: extended serialized ChartVariable object (dict).
  """
  return dict(
    **standard(cv),
    **input_serializer.in_variable(cv.input_model)
  )



import base64
import json
from typing import List, Dict

from flask import Blueprint, jsonify

from kama_sdk.model.base.model import Model
from kama_sdk.model.concern.concern_view_adapter import ConcernViewAdapter
from kama_sdk.model.concern.concern_detail_adapter import ConcernDetailAdapter
from kama_sdk.model.concern.concern_set import ConcernSet
from kama_sdk.model.concern.concern_super_set import ConcernSuperSet
from kama_sdk.serializers import concern_ser

controller = Blueprint('concerns_controller', __name__)

BASE_PATH = '/api/concerns'
DETAIL_PATH = f'{BASE_PATH}/details/<b64_enc_reconstructor>'


@controller.route(f'{BASE_PATH}/super_sets/<super_set_id>')
def get_category_sets(super_set_id):
  super_set: ConcernSuperSet = ConcernSuperSet.inflate(super_set_id)
  if super_set:
    sets: List[ConcernSet] = super_set.concern_sets
    serialize = lambda s: concern_ser.serialize_set_meta(s)
    return jsonify(data=list(map(serialize, sets)))
  else:
    return jsonify({'error': f"no such category {super_set_id}"}), 400


@controller.route(f'{BASE_PATH}/sets/<concern_set_id>')
def get_set_meta(concern_set_id):
  concern_set: ConcernSet = ConcernSet.inflate(concern_set_id)
  if concern_set:
    meta = concern_ser.serialize_set_meta(concern_set, True)
    return jsonify(data=meta)
  else:
    return jsonify({'error': f"no such set {concern_set_id}"}), 400


@controller.route(f'{BASE_PATH}/sets/<concern_set_id>/data')
def get_set_data(concern_set_id):
  concern_set: ConcernSet = Model.inflate(concern_set_id)
  if concern_set:
    data = concern_set.one_shot_compute()
    return jsonify(data=data)
  else:
    return jsonify({'error': f"no such set {concern_set_id}"}), 400


@controller.route(f'{BASE_PATH}/cards/<b64_enc_str>')
def get_card(b64_enc_str):
  reconstructor = base64.b64decode(b64_enc_str)
  if adapter := ConcernDetailAdapter.reconstruct(reconstructor):
    data = adapter.compute()
    return jsonify(data=data)
  else:
    return jsonify({'error': f"no adapter", 'r': reconstructor}), 400


@controller.route(DETAIL_PATH)
def get_detail_meta(b64_enc_reconstructor):
  if adapter := get_adapter(b64_enc_reconstructor):
    serialized = concern_ser.ser_detail_meta(adapter)
    return jsonify(data=serialized)
  else:
    return jsonify({'error': f"no adapter"}), 400


@controller.route(f'{DETAIL_PATH}/data/<group_type>/<group_id>')
def get_detail_group_value(b64_enc_reconstructor, group_type, group_id):
  adapter: ConcernDetailAdapter = get_adapter(b64_enc_reconstructor)
  if adapter:
    data = adapter.compute_group_data(group_type, group_id)
    return jsonify(data=data)
  else:
    return jsonify({'error': f"no adapter"}), 400


def decode_seed(b64_enc_str: str) -> Dict:
  utc_str = base64.b64decode(b64_enc_str)
  return json.loads(utc_str)


def get_adapter(b64_enc_constructor: str):
  utf_reconstructor = base64.b64decode(b64_enc_constructor)
  reconstructor = json.loads(utf_reconstructor)
  return ConcernViewAdapter.reconstruct(reconstructor)



import json
from typing import Dict


def ser_config_backup(config: Dict) -> Dict:
  data = config.get('data') or {}
  then_ktea = json.loads(data.get('ktea') or '{}')
  then_wiz = json.loads(data.get('wiz') or '{}')

  return dict(
    id=str(config.get('_id')),
    name=config.get('name'),
    event_type=config.get('event_type'),
    then_status=data.get('status'),
    then_ktea=then_ktea,
    then_wiz=then_wiz,
    timestamp=config.get('timestamp')
  )


def ser_config_backup_full(config: Dict) -> Dict:
  return dict(
    **ser_config_backup(config),
    data=config.get('data', {})
  )



from typing import List, TypeVar, Optional, Any

from werkzeug.utils import cached_property

from kama_sdk.core.core.types import PredEval
from kama_sdk.core.core.utils import any2bool, perr
from kama_sdk.model.base.model import Model
from kama_sdk.model.input.generic_input import GenericInput
from kama_sdk.model.supplier.predicate import predicate as predicate_module
from kama_sdk.model.supplier.predicate.predicate import Predicate

T = TypeVar('T', bound='GenericVariable')

class GenericVariable(Model):

  @cached_property
  def default_value(self) -> Optional[Any]:
    default_value = self.get_prop(DEFAULT_VALUE_KEY)
    if default_value is None:
      if self.input_model:
        default_value = self.input_model.compute_inferred_default()
    return default_value

  @cached_property
  def input_model(self) -> GenericInput:
    return self.inflate_child(
      GenericInput,
      prop=INPUT_MODEL_KEY,
      safely=True
    ) or GenericInput({})

  def compute_decoration(self) -> Optional[str]:
    return self.get_prop(DECORATION_KEY)

  def validation_predicates(self, value: Any) -> List[Predicate]:
    value = self.sanitize_for_validation(value)
    return self.inflate_children(
      Predicate,
      prop=VALIDATION_PREDS_KEY,
      resolve_kod=False,
      patches={
        predicate_module.CHALLENGE_KEY: value
      }
    )

  def validate(self, value: Any) -> PredEval:
    try:
      patched_predicates = self.validation_predicates(value)
      for predicate in patched_predicates:
        if not any2bool(predicate.resolve()):
          return PredEval(
            predicate_id=predicate.id(),
            met=False,
            reason=predicate.reason(),
            tone=predicate.tone()
          )
    except:
      perr(self, f"inflate/resolve failed for validators", True)
      return PredEval(
        met=False,
        tone='error',
        reason='error loading or running validator(s)'
      )
    return PredEval(
        met=True,
        tone='',
        reason=''
      )

  def current_or_default_value(self):
    return self.default_value

  def sanitize_for_validation(self, value: Any) -> Any:
    return self.input_model.sanitize_for_validation(value)


DEFAULT_VALUE_KEY = 'default'
DECORATION_KEY = 'decorator'
INPUT_MODEL_KEY = 'input'
VALIDATION_PREDS_KEY = 'validators'


COPYABLE_KEYS = [
  DEFAULT_VALUE_KEY,
  DECORATION_KEY,
  INPUT_MODEL_KEY,
  VALIDATION_PREDS_KEY
]



from typing import List

from werkzeug.utils import cached_property

from kama_sdk.model.base.model import Model
from kama_sdk.model.concern.concern_set import ConcernSet


class ConcernSuperSet(Model):

  @cached_property
  def concern_sets(self) -> List[ConcernSet]:
    return self.inflate_children(
      ConcernSet,
      prop=SETS_KEY,
    )

  @cached_property
  def slug(self) -> str:
    return self.resolve_prop(LABEL_KEY, lookback=0) or self.id()


LABEL_KEY = 'label'
SETS_KEY = 'sets'



import os

import redis
from rq import Worker, Connection

from kama_sdk.core.core import consts

redis_url = os.getenv('WORK_REDIS_URL', 'redis://localhost:6379')

conn = redis.from_url(redis_url)

def start_main():
  _start(consts.MAIN_WORKER)


def start_telem():
  _start(consts.TELEM_WORKER)


def _start(queue_name: str):
  with Connection(conn):
    worker = Worker(
      queues=[queue_name],
      connection=conn
    )
    worker.work()



import sys
import time

from kama_sdk.core.core import utils, job_client
from kama_sdk.core.core.config_man import coerce_ns
from kama_sdk.core.core.types import ActionStatusDict
from kama_sdk.model.operation.operation_run_simulator import OperationRunSimulator
from kama_sdk.model.operation.operation_state import OperationState


def start():
  simulator: OperationRunSimulator = read_simulator()
  operation = simulator.operation
  coerce_ns(simulator.force_ns)
  state = OperationState(utils.rand_str(), operation.id())

  print(f"Operation[{operation.title or operation.id()}]")

  stage_ind = 0
  step_id = None

  if simulator.run_preflight:
    print("   Preflight Checks: ")
    preflight_action = operation.preflight_action_config()
    job_id = job_client.enqueue_action(preflight_action)
    status = 'running'
    while status == 'running':
      progress = job_client.job_progress(job_id)
      status = job_client.ternary_job_status(job_id)
      if progress:
        print_progress(progress)
      else:
        print("        No progress :/")
      time.sleep(3)
    print("   Preflight Checks complete")

  while simulator.stage_at(stage_ind) is not None:
    stage = simulator.stage_at(stage_ind)
    print(f"  Stage [{stage.title}]:")
    if step_id is None:
      step_id = stage.steps[0].id() if len(stage.steps) > 0 else 'done'
    while not step_id == 'done':
      step = stage.step(step_id)
      print(f"    Step [{step.title}]:")
      step_input = simulator.step_input(stage_ind, step_id)
      if simulator.validate_input:
        print("     Validation")
        for input_key, input_val in step_input.items():
          validation = step.validate_field(input_key, input_val, state)
          met, tone = validation['met'], validation['tone']
          pretty = validation_emoji(met, tone)
          print(f"        {input_key}={input_val} => {pretty}")
          if tone == 'error' and not met:
            return "HALT VALIDATION FAILED"
        print("     Validation complete")
      step_state = state.gen_step_state(step)
      step.run(step_input, step_state)
      step.compute_status(step_state)
      print("     Run")
      while step_state.status == 'running':
        job_id = step_state.job_id
        progress = job_client.job_progress(job_id)
        print_progress(progress)
        time.sleep(3)
        step.compute_status(step_state)
      print("     Run Complete")
      step_id = stage.next_step_id(step, state)
      print("   Step complete")
    print(" Stage complete")
    stage_ind += 1
  print("Operation complete")


def print_progress(progress: ActionStatusDict):
  sub_items = progress.get('sub_items', [])
  for item in sub_items:
    if not item['status'] == 'idle':
      base = f"        {item['title']}"
      print(f"{base}: {status_emoji(item['status'])}")
      for sub_sub_item in item['sub_items']:
        base = f"          {sub_sub_item['title']}"
        print(f"{base}: {status_emoji(sub_sub_item['status'])}")
      tone = item.get('data', {}).get('tone')
      reason = item.get('data', {}).get('reason')
      if tone and reason:
        print(f"          {tone}: {reason}")


def read_simulator() -> OperationRunSimulator:
  _id = sys.argv[2] if len(sys.argv) >= 3 else None
  return OperationRunSimulator.inflate(_id)


def status_emoji(charge):
  if charge == 'running':
    return '🏃‍️'
  elif charge == 'positive':
    return '✅'
  elif charge == 'negative':
    return " 🚨"
  elif charge == 'idle':
    return "✋"
  else:
    return "⁉"


def validation_emoji(met, tone):
  if met:
    return "✔"
  else:
    if tone == 'warning':
      return "⚠"
    elif tone == 'error':
      return "🚨💥💣💣💣💣🧨🥀🚨🚨⚠⚠"
    else:
      return "⁉"



from datetime import datetime
from typing import List

import humanize
from werkzeug.utils import cached_property

from kama_sdk.core.core.types import PromSeriesDataPoint, SimpleSeriesSummary
from kama_sdk.model.glance.glance import Glance


class TimeSeriesGlance(Glance):

  @cached_property
  def series_summary(self) -> SimpleSeriesSummary:
    return self.get_prop(DATA_KEY, None)

  @cached_property
  def time_series(self) -> List[PromSeriesDataPoint]:
    return self.series_summary['series']

  @cached_property
  def view_type(self) -> str:
    return 'line_chart'

  @cached_property
  def info(self):
    if self.series_summary and len(self.time_series) > 0:
      ts0 = self.time_series[0]['timestamp']
      delta_str = humanize.naturaltime(datetime.fromisoformat(ts0))
      return f"{delta_str} - now"
    else:
      return "Not enough data"

  @cached_property
  def reducer_type(self) -> str:
    return self.get_prop(REDUCER_FUNC_KEY, 'last')

  def gen_legend(self):
    if self.series_summary:
      return {
        **super().gen_legend(),
        'direction': self.series_summary['direction'],
        'good_direction': self.series_summary['good_direction']
      }
    else:
      return {}

  def content_spec(self):
    return {
      'timeseries': self.series_summary['humanized_series'],
      'value': self.series_summary['summary_value'],
      'xKey': 'timestamp',
      'yKey': 'value'
    }


DATA_KEY = 'data'
REDUCER_FUNC_KEY = 'reducer'
HUMANIZER_KEY = 'humanizer'
GOOD_DIRECTION_KEY = 'good_direction'



from typing import List, Dict

from werkzeug.utils import cached_property

from kama_sdk.core.core.types import ConcernAttrMeta
from kama_sdk.model.concern import concern_attr_adapter
from kama_sdk.model.concern.concern import Concern
from kama_sdk.model.concern.concern_attr_adapter import ConcernAttrAdapter
from kama_sdk.model.concern.concern_set import ConcernSet


class ConcernTable(ConcernSet):

  @cached_property
  def concern_shell(self) -> Concern:
    shell = self.inflate_child(
      Concern,
      prop=CONCERN_SHELL_KEY,
      safely=True
    )
    return shell or Concern({})

  def concern_constructors(self) -> List[Dict]:
    return self.get_prop(CONCERN_CONSTRUCTORS_KEY) or []

  @cached_property
  def is_one_shot(self) -> bool:
    return True

  @cached_property
  def attr_adapters(self) -> List[ConcernAttrAdapter]:
    return self.inflate_children(
      ConcernAttrAdapter,
      prop=COLUMNS_KEY
    )

  @cached_property
  def column_definitions(self) -> List[ConcernAttrMeta]:
    return [a.as_meta() for a in self.attr_adapters]

  def concern_constructor_to_row(self, constructor:  Dict) -> Dict:
    concern_instance: Concern = self.concern_shell.clone().patch(constructor)
    row_data = {}
    for adapter_shell in self.attr_adapters:
      adapter = adapter_shell.clone().patch({
        concern_attr_adapter.CONCERN_KEY: concern_instance,
        concern_attr_adapter.SET_ID_KEY: self.id()
      })
      row_data[adapter.label] = adapter.compute_value()
    return row_data

  def one_shot_compute(self) -> List[Dict]:
    constructors = self.concern_constructors()
    rows = []
    for constructor in constructors:
      row = self.concern_constructor_to_row(constructor)
      rows.append(row)
    return rows

  @classmethod
  def view_type(cls) -> str:
    return 'table'


CONCERN_CONSTRUCTORS_KEY = 'concern_constructors'
COLUMNS_KEY = 'columns'
CONCERN_SHELL_KEY = "concern_shell"



from typing import Optional, Dict

import requests

from kama_sdk.core.core import utils
from kama_sdk.core.core.config_man import config_man

api_path = '/cli'


def backend_host() -> Optional[str]:
  if config_man.is_training_mode():
    print("[kama_sdk:hub_client] cannot call API with prototype")
    return None
  if utils.is_dev() or utils.is_test():
    if utils.is_in_cluster():
      return "http://necthub.com.ngrok.io"
    else:
      return "http://localhost:3000"
  else:
    return 'https://api.codesdk.com'


def post(endpoint, payload) -> requests.Response:
  url = f'{backend_host()}{api_path}{endpoint}'
  print(f"[kama_sdk:hub_client] post {url}")
  return requests.post(
    url,
    json=payload,
    headers=gen_headers()
  )


def patch(endpoint, payload) -> requests.Response:
  url = f'{backend_host()}{api_path}{endpoint}'
  print(f"[kama_sdk:hub_client] patch {url}")
  return requests.patch(
    url,
    json=payload,
    headers=gen_headers()
  )


def get(endpoint) -> requests.Response:
  url = f'{backend_host()}{api_path}{endpoint}'
  print(f"[kama_sdk:hub_client] get {url}")
  return requests.get(
    url,
    headers=gen_headers()
  )

def gen_headers() -> Dict:
  access_token = config_man.install_token()
  return {
    'Token': access_token
  }



from kama_sdk.model.operation import step

step.Step._partition_vars_by_field.__doc__ = """
Partitions new assigns into 3 buckets (chart, inline, state) and
merges with existing assigns in each bucket.
:param assigns: new assigns to be partitioned.
:param prev_state: state from which to collect existing assigns.
:return: tuple of all chart assigns, partitioned and merged.
"""


step.Step.fields.__doc__ = """
Finds the Field by key and inflates (instantiates) into a Field instance.
:param key: identifier for desired Field.
:return: Field instance.
"""


step.Step.has_explicit_next.__doc__ = """
Checks if the current step has an explicit next step.
:return: True if it does, otherwise False.
"""


step.Step.next_step_id.__doc__ = """
Returns the id of the next step, be it explicit or an if-then-else type step.
:param values: if-then-else values, if necessary.
:return: string containing next step.
"""



from typing import Dict, List, Any, Optional
from werkzeug.utils import cached_property

from kama_sdk.core.core.types import InputOption
from kama_sdk.model.base.model import Model


class GenericInput(Model):

  @cached_property
  def options(self) -> List[InputOption]:
    return self.get_prop(OPTIONS_KEY, [])

  def compute_inferred_default(self) -> Optional[Any]:
    if len(self.options) > 0:
      return self.options[0].get('id')
    return None

  def serialize_options(self) -> List[InputOption]:
    return self.options

  def extras(self) -> Dict[str, any]:
    return {}

  @staticmethod
  def sanitize_for_validation(value: Any) -> Any:
    return value


OPTIONS_KEY = 'options'



import traceback
from typing import Any, List, Optional, Dict

import yaml
from typing_extensions import TypedDict

from kama_sdk.core.core import utils
from kama_sdk.model.base import model
from kama_sdk.model.base.model import models_man, ID_REFERENCE_PREFIX, KIND_REFERENCE_PREFIX
from termcolor import colored, cprint


class Violation(TypedDict):
  severity: str
  id: Optional[str]
  type: str
  message: Optional[str]
  desc: Optional[Dict]


def run_all():
  try:
    violations = []
    violations.extend(perform_top_level_validations())
    render_violations(violations)
  except:
    print("static check failed")
    if not utils.is_prod():
      print(traceback.format_exc())


def render_violations(violations: List[Violation]):
  if len(violations) > 0:
    cprint(f"\n{'*' * 53}", 'magenta')
    cprint(f"                  {len(violations)} PROBLEMS FOUND", attrs=['bold'])
    cprint(f"{'*' * 53}\n", 'magenta')
    for violation in violations:
      render_violation(violation)
    print("")


def render_violation(violation: Violation):
  _id = violation['id']
  severity = violation['severity']
  message = violation['message']
  _type = violation['type']
  desc = violation['desc']

  color = "red" if severity == 'error' else 'magenta'

  cprint(f"\n{severity.upper()} {_type}", color, attrs=['bold'])

  if desc:
    cprint("For")
    cprint(f"{yaml.dump(desc, indent=5)}")
  else:
    print(f"For {colored(_id, attrs=['underline'])}")

  if message:
    cprint(message, color)

def descriptor_id_refs(descriptor: Any) -> List[str]:
  id_refs = []
  if type(descriptor) == dict:
    for sub_tree in descriptor.values():
      new_refs = descriptor_id_refs(sub_tree)
      id_refs.extend(new_refs)
  elif type(descriptor) == list:
    for item in descriptor:
      new_refs = descriptor_id_refs(item)
      id_refs.extend(new_refs)
  elif type(descriptor) == str:
    if descriptor.startswith(ID_REFERENCE_PREFIX):
      id_refs.append(descriptor)
    elif descriptor.startswith(KIND_REFERENCE_PREFIX):
      id_refs.append(descriptor)
  return id_refs


def perform_top_level_validations() -> List[Violation]:
  violations: List[Violation] = []
  descriptors = models_man.descriptors()
  for descriptor in descriptors:
    if ref_violation := ensure_top_level_to_child_refs(descriptor):
      violations.append(ref_violation)

    if id_violation := ensure_top_level_id_present(descriptor):
      violations.append(id_violation)

    if kind_violation := ensure_kind_class_exists(descriptor):
      violations.append(kind_violation)

  return violations


def ensure_top_level_to_child_refs(descriptor: Dict) -> Optional[Violation]:
  refs = descriptor_id_refs(descriptor)
  for ref in refs:
    can_handle = True
    referee = None

    un_prefixed_ref = ref.split("::")[-1]
    if ref.startswith(ID_REFERENCE_PREFIX):
      referee = model.find_any_config_by_id(un_prefixed_ref)
    elif ref.startswith(KIND_REFERENCE_PREFIX):
      referee = model.find_any_class_by_name(un_prefixed_ref)
    else:
      can_handle = False

    if can_handle and not referee:
      return Violation(
        severity='error',
        id=descriptor.get('id'),
        type='Broken reference',
        message=f"Target --> {un_prefixed_ref} not found",
        desc=None
      )


def ensure_top_level_id_present(descriptor: Dict) -> Optional[Violation]:
  _id = descriptor.get('id')
  if not _id:
    return Violation(
      severity='error',
      id=None,
      type="Missing top-level property 'id'",
      message=None,
      desc=descriptor
    )


def ensure_kind_class_exists(descriptor: Dict) -> Optional[Violation]:
  if kind := descriptor.get('kind'):
    if not model.find_any_class_by_name(kind):
      return Violation(
        severity='error',
        type='No class for kind',
        id=descriptor.get('id'),
        message=f"Expected '{kind}' to exist but it does not",
        desc=None if descriptor.get('id') else descriptor
      )
  else:
    return Violation(
      severity='error',
      type="Missing top-level property 'kind'",
      id=descriptor.get('id'),
      message=None,
      desc=None if descriptor.get('id') else descriptor
    )


def ensure_single_publisher_definitions():
  pass


def ensure_inherit_id_refs():
  pass


def ensure_kind_refs():
  pass



from typing import Dict, List

from kama_sdk.core.core.types import ErrCapture
from kama_sdk.core.ktea.ktea_client import KteaClient
from kama_sdk.model.action.base.action import Action
from kama_sdk.model.action.base.action_errors import FatalActionError


class KubectlDryRunAction(Action):

  def res_descs(self) -> List[Dict]:
    return self.get_prop('res_descs', [])

  def perform(self) -> None:
    success, logs = KteaClient.kubectl_dry_run(self.res_descs())
    self.add_logs(logs)
    raise_on_dry_run_err(success, logs)


def raise_on_dry_run_err(success: bool, logs):
  if not success:
    raise FatalActionError(ErrCapture(
      type='kubectl_dry_run_failed',
      name=f"manifest was rejected",
      reason='kubectl dry_run failed for one or more resource.',
      logs=logs
    ))



from typing import Dict

from kama_sdk.model.adapter.resource.resource_table import ResourceTable
from kama_sdk.model.adapter.resource.resource_table_column import ResourceTableColumn


def serialize_col_meta(column: ResourceTableColumn) -> Dict:
  return {
    'label': column.attribute_label(),
    'title': column.title
  }


def serialize_meta(table: ResourceTable) -> Dict:
  return {
    'id': table.id(),
    'title': table.title,
    'columns': list(map(serialize_col_meta, table.columns())),
    'searchable_labels': table.searchable_labels
  }



from typing import Dict, Optional

from k8kat.res.config_map.kat_map import KatMap
from werkzeug.utils import cached_property

from kama_sdk.core.core.config_man import config_man

cmap_field = 'sharing_prefs'
upload_telem_key = 'upload_telem'


class TelemPerms:
  """This class is responsible for Telemetry screen in the Pre-installation."""

  @cached_property
  def master_cmap(self) -> KatMap:
    """
    Uses Kteai client to return the ConfigMap.
    :return: ConfigMap object.
    """
    return config_man.load_master_cmap()

  def user_perms(self) -> Dict:
    """
    Gets the default "sharing prefs" value from the ConfigMap.
    :return: dict with sharing prefs.
    """
    kat_map = self.master_cmap
    return kat_map.jget(cmap_field, {}) if kat_map else {}

  def patch(self, values: Dict):
    """
    Inserts the passed values into the user_perms dictionary.
    :param values: key-value pairs to be isnerted.
    """
    values = {k: v for k, v in values.items() if v is not None}
    if self.master_cmap:
      merged = {**self.user_perms(), **values}
      self.master_cmap.jpatch(merged, cmap_field, merge=False)

  def can_sync_telem(self) -> bool:
    """
    Gets the "upload telem" key from user_perms and converts it to a boolean.
    This defines the permission to sync telem information with the database.
    :return:
    """
    return x_to_bool(self.user_perms().get(upload_telem_key))

  def can_share_prop(self, prop: str) -> bool:
    """
    Gets the value of the category that matches the passed property from
    user_perms and converts it to boolean.
    :param prop: property to find the category.
    :return: True if value is True, otherwise False.
    """
    category = find_prop_category(prop)
    return x_to_bool(category and self.user_perms().get(category))


def x_to_bool(raw) -> bool:
  """
  Converts string / None / bool to bool.
  :param raw: input to be converted.
  :return: boolean value.
  """
  if raw is None:
    return False
  elif type(raw) == bool:
    return raw
  elif type(raw) == str:
    return raw.lower() == 'true'


def find_prop_category(prop) -> Optional[str]:
  """
  Returns category based on category property.
  :param prop: category property.
  :return: category.
  """
  for category, category_props in category_props_mapping.items():
    if prop in category_props:
      return category
  return None


category_props_mapping = {
  'telem.persist': [],
  'telem.upload': [],
  'telem.share_vendor': [],
  'operations.metadata': [
    'operation_outcome.operation_id',
    'operation_outcome.step_outcomes',
    'operation_outcome.step_outcomes.stage_id',
    'operation_outcome.step_outcomes.step_id',
    'operation_outcome.step_outcomes.started_at',
    'operation_outcome.step_outcomes.terminated_at',
    'operation_outcome.step_outcomes.job_logs',
    'operation_outcome.step_outcomes.commit_outcome',
    'operation_outcome.step_outcomes.exit_condition_outcomes',
    'operation_outcome.step_outcomes.exit_condition_outcomes.condition_id',
    'operation_outcome.step_outcomes.exit_condition_outcomes.condition_met',
    'operation_outcome.step_outcomes.exit_condition_outcomes.reason'
  ],
  'operations.assignments': [
    'operation_outcome.step_outcomes.chart_assigns',
    'operation_outcome.step_outcomes.state_assigns',
  ],
  'operations.asd': [

  ]
}



from typing import List

from kama_sdk.core.core import utils, consts
from kama_sdk.core.core.types import KAO
from kama_sdk.model.base import model
from kama_sdk.model.supplier.ext.biz import resources_supplier as res_sup
from kama_sdk.model.supplier.ext.biz import resource_selector as res_sel
from kama_sdk.model.supplier.base import supplier as sup
from kama_sdk.model.supplier.predicate.predicate import Predicate
from kama_sdk.model.supplier.predicate import predicate as pred


def outkomes2preds(outkomes: List[KAO]) -> List[Predicate]:
  predicates = dict(positive=[], negative=[])
  for ktl_outcome in outkomes:
    if not ktl_outcome['verb'] == 'unchanged':
      for charge in [consts.pos, consts.neg]:
        predicate = outkome2charged_pred(ktl_outcome, charge)
        predicates[charge].append(predicate)
  return utils.flatten(predicates.values())


def outkome2charged_pred(ktl_out: KAO, charge):
  from kama_sdk.model.supplier.ext.biz.resources_supplier import ResourcesSupplier

  selector = {
    res_sel.RES_KIND_KEY: ktl_out['kind'],
    res_sel.RES_NAME_KEY: ktl_out['name'],
    'api_group': ktl_out['api_group']
  }

  return Predicate({
    model.KIND_KEY: Predicate.__name__,
    model.ID_KEY: f"{ktl_out['kind']}/{ktl_out['name']}-{charge}",
    model.TITLE_KEY: f"{ktl_out['kind']}/{ktl_out['name']} is {charge}",
    pred.CHECK_AGAINST_KEY: charge,
    pred.IS_OPTIMISTIC_KEY: charge == consts.pos,
    pred.OPERATOR_KEY: '=',
    pred.ON_MANY_KEY: 'each_true',
    pred.CHALLENGE_KEY: {
      model.KIND_KEY: ResourcesSupplier.__name__,
      sup.OUTPUT_FMT_KEY: 'ternary_status',
      sup.SERIALIZER_KEY: 'legacy',
      sup.IS_MANY_KEY: True,
      res_sup.RESOURCE_SELECTOR_KEY: selector
    },
    pred.ERROR_EXTRAS_KEY: dict(
      resource_signature=dict(kind=ktl_out['kind'], name=ktl_out['name']),
      resource={
        model.KIND_KEY: ResourcesSupplier.__name__,
        res_sup.RESOURCE_SELECTOR_KEY: selector,
        sup.IS_MANY_KEY: False
      }
    )})



from typing import List, TypeVar, Dict

from werkzeug.utils import cached_property

from kama_sdk.core.core import consts
from kama_sdk.core.core.utils import any2bool
from kama_sdk.model.base.model import Model, INFO_KEY, TITLE_KEY
from kama_sdk.model.variable import generic_variable
from kama_sdk.model.variable.generic_variable import GenericVariable

T = TypeVar('T', bound='Field')

class Field(Model):

  @cached_property
  def variable(self) -> GenericVariable:
    """
    Inflate child variable using 'variable' KoD. If not present,
    construct synthetic GenericVariable using own config.
    @return: GenericVariable instance
    """
    return self.inflate_child(
      GenericVariable,
      prop=VARIABLE_KEY,
      safely=True
    ) or self.inflate_child(
      GenericVariable,
      kod=self.variable_bound_config_subset()
    )

  @cached_property
  def title(self) -> str:
    if value := self.resolve_prop(TITLE_KEY, lookback=0):
      return value
    else:
      return self.variable.title

  @cached_property
  def info(self) -> str:
    if value := self.resolve_prop(INFO_KEY, lookback=0):
      return value
    else:
      return self.variable.info

  @cached_property
  def target(self):
    return self.get_prop(TARGET_KEY, consts.DEFAULT_TARGET)

  def decorate(self):
    pass

  def variable_bound_config_subset(self):
    pool = self.config.items()
    relevant = generic_variable.COPYABLE_KEYS
    config = {k: v for k, v in pool if k in relevant}
    return config

  def compute_visibility(self) -> bool:
    result = self.get_prop(VISIBLE_KEY, True)
    return any2bool(result)

  def serialize_options(self) -> List[Dict]:
    return self.variable.input_model.serialize_options()

  def is_manifest_bound(self) -> bool:
    return self.is_chart_var() or self.is_inline_chart_var()

  def is_inline_chart_var(self) -> bool:
    return self.target == consts.TARGET_INLIN

  def is_chart_var(self) -> bool:
    return self.target == consts.TARGET_STANDARD

  def is_state_var(self) -> bool:
    return self.target == consts.TARGET_STATE


TARGET_KEY = 'target'
VARIABLE_KEY = 'variable'
VISIBLE_KEY = 'visible'




from datetime import datetime, timedelta
from typing import Dict, Optional, Union

from werkzeug.utils import cached_property

from kama_sdk.core.core import prom_api_client
from kama_sdk.core.core.types import PromMatrix, PromVector
from kama_sdk.model.supplier.base.supplier import Supplier


class PromDataSupplier(Supplier):

  @cached_property
  def step(self) -> str:
    return self.get_prop(STEP_KEY, '1h')

  @cached_property
  def t0(self) -> datetime:
    offset = self.get_prop(T0_OFFSET_KEY, {'days': 7})
    return parse_from_now(offset)

  @cached_property
  def tn(self) -> datetime:
    offset = self.get_prop(TN_OFFSET_KEY, {})
    return parse_from_now(offset)

  @cached_property
  def serializer_type(self) -> str:
    return self.get_prop('serializer', 'legacy')

  @cached_property
  def _type(self) -> str:
    return self.resolve_prop('type', backup='matrix', lookback=0)

  def resolve(self) -> Union[PromMatrix, PromVector]:
    if self._type == 'matrix':
      response = self.fetch_matrix()
    elif self._type == 'vector':
      response = self.fetch_vector()
    else:
      print(f"[kama_sdk:prom_supplier] bad req type {self._type}")
      response = None

    return response

  def fetch_matrix(self) -> Optional[PromMatrix]:
    prom_data = prom_api_client.compute_matrix(
      self.source_data(),
      self.step,
      self.t0,
      self.tn
    )
    # print("RAW")
    # print(prom_data)
    return prom_data['result'] if prom_data else None

  def fetch_vector(self) -> Optional[PromVector]:
    prom_data = prom_api_client.compute_vector(
      self.source_data(),
      self.tn
    )
    return prom_data['result'] if prom_data else None

def parse_from_now(expr: Dict) -> datetime:
  difference = {k: int(v) for k, v in expr.items()}
  return datetime.now() - timedelta(**difference)


STEP_KEY = 'step'
T0_OFFSET_KEY = 't0_offset'
TN_OFFSET_KEY = 'tn_offset'



from werkzeug.utils import cached_property

from kama_sdk.core.core.types import EndpointDict
from kama_sdk.model.base.model import Model


class SiteConcern(Concern):

  def endpoint_bundle(self) -> EndpointDict:
    pass

  def is_working(self) -> bool:
    pass

  def status(self) -> str:
    return "positive" if self.is_working() else "negative"

  def logo_url(self) -> str:
    return self.get_prop('logo_url')


class SiteCard(Model):

  @cached_property
  def concern(self) -> SiteConcern:
    return self.inflate_child(
      SiteConcern,
      prop='concern'
    )

  def header_desc(self):
    site_name = self.concern.title
    site_info = self.concern.info
    status = self.concern.status()
    col = "pleasant" if self.concern.is_working() else "ruby"
    logo_url = self.concern.logo_url()
    return _header_desc(site_name, status, col, logo_url, site_info)


def _header_desc(name, status, col, img, info):
  return {
    'type': 'ThreePartHeader',
    'graphic': {
      'type': 'Image',
      'uri': img
    },
    'title': {
      'type': 'Line',
      'elements': [
        {
          'type': 'Text',
          'text': name
        },
        {
          'type': 'Tag',
          'text': status,
          'style': {'bkgCol': col}
        }
      ]
    },
    'subtitle': {
      'type': 'Text',
      'text': info
    }
  }




import json
import traceback
from datetime import datetime
from typing import Dict, Any, List, Optional, TypeVar

from inflection import underscore
from rq import get_current_job
from rq.job import Job
from werkzeug.utils import cached_property

from kama_sdk.core.core import utils, consts, job_client
from kama_sdk.core.core.types import ActionStatusDict, ErrCapture, EventCapture
from kama_sdk.core.telem import telem_man
from kama_sdk.model.action.base.action_errors import ActionErrOrException, ActionError
from kama_sdk.model.base.model import Model

T = TypeVar('T', bound='Action')

class Action(Model):

  status: str
  err_capture: Optional[ErrCapture]
  telem_vid: str
  logs: List[str]

  def __init__(self, config: Dict):
    super().__init__(config)
    self.status = 'idle'
    self.logs = []
    self.err_capture = None
    self.telem_vid = utils.gen_uuid()

  def run(self) -> Any:
    outcome: Optional[Any]
    exception: Optional[Exception]
    err_capture: Optional[ErrCapture] = None
    should_raise: bool = False

    try:
      self.set_running()
      outcome = self.perform()
      exception = None
    except Exception as _exception:
      print("WTF")
      print(traceback.format_exc())
      outcome = None
      exception = _exception

    if exception:
      status = consts.neg
      err_capture = exception2capture(self.telem_vid, exception)
      should_raise = err_capture['fatal']
      self.add_logs(err_capture.get('logs'))
    else:
      status = consts.pos

    self.err_capture = err_capture
    self.set_status(status)
    self.handle_telem()

    if should_raise:
      raise ActionError(err_capture, is_original=False)

    return outcome

  def parent_telem_event_id(self):
    return self.get_prop(KEY_PARENT_EVENT_VID)

  @cached_property
  def debug_prop_names(self) -> List[str]:
    return self.get_prop(DEBUG_PROPS_KEY, [])

  @cached_property
  def telem_enabled(self) -> List[str]:
    return self.get_prop(TELEM_ENABLED_KEY, True)

  def gen_backup_event_name(self):
    return underscore(self.__class__.__name__)

  def handle_telem(self):
    job_client.enqueue_or_run_telem_func(
      not utils.is_test(),
      telem_man.store_event,
      self.gen_telem_bundle()
    )

    if self.err_capture and self.err_capture['is_original']:
      job_client.enqueue_or_run_telem_func(
        not utils.is_test(),
        telem_man.store_error,
        self.err_capture
      )

    job_client.enqueue_or_run_telem_func(
      not utils.is_test(),
      telem_man.upload_events_and_errors,
    )

  def gen_telem_bundle(self) -> EventCapture:
    parent_action = self.parent_action()
    parent_vid = parent_action.telem_vid if parent_action else None
    return EventCapture(
      vid=self.telem_vid,
      parent_vid=parent_vid,
      type=ACTION_EVENT_TYPE,
      name=self.id() or self.gen_backup_event_name(),
      status=self.status,
      logs=self.logs,
      occurred_at=str(datetime.now())
    )

  def set_running(self):
    self.set_status(consts.rng)

  def set_positive(self):
    self.set_status(consts.pos)

  def set_negative(self):
    self.set_status(consts.neg)

  def set_status(self, status):
    assert status in consts.statuses
    self.status = status
    self.notify_job()

  def perform(self) -> Optional[Dict]:
    raise NotImplementedError

  def add_logs(self, new_logs: Optional[List[str]]) -> None:
    new_logs = new_logs or []
    self.logs = [*self.logs, *new_logs]

  def parent_action(self) -> Optional[T]:
    if self.parent:
      if issubclass(self.parent.__class__, Action):
        return self.parent
    return None

  def am_sub_action(self) -> bool:
    return issubclass(self.parent.__class__, Action)

  def am_root_action(self):
    return not self.am_sub_action()

  def find_root_action(self) -> Optional:
    if self.am_root_action():
      return self
    elif parent_action := self.parent_action():
      return parent_action.find_root_action()
    else:
      return None

  def notify_job(self):
    job: Job = get_current_job()
    if job:
      action_root = self.find_root_action()
      if action_root:
        progress_bundle = action_root.serialize_progress()
        job.meta['progress'] = json.dumps(progress_bundle)
        job.save_meta()
      else:
        print(f"[action:{self.id}] danger root not found")

  def gen_debug_dump(self) -> Dict:
    bundle = {}
    for prop_name in self.debug_prop_names:
      try:
        bundle[prop_name] = self.get_prop(prop_name)
      except RuntimeError as e:
        bundle[prop_name] = f"[error] {str(e)}"
    return bundle

  def serialize_progress(self) -> ActionStatusDict:
    return dict(
      id=self.id(),
      title=self.title,
      info=self.info,
      status=self.status,
      sub_items=[],
      logs=self.logs,
      debug=self.gen_debug_dump(),
      error=err2client_facing(self.err_capture)
    )

  @cached_property
  def expected_run_args(self) -> List[str]:
    return self.get_prop(EXPECTED_RUN_ARGS_KEY, [])


def err2client_facing(err_capt: ErrCapture) -> Optional[Dict]:
  if err_capt:
    return dict(
      fatal=err_capt.get('fatal', True),
      type=err_capt.get('type') or 'unknown_type',
      reason=err_capt.get('reason') or 'Unknown Reason',
      logs=err_capt.get('logs', [])
    )
  else:
    return None


def exception2capture(vid: str, exception: ActionErrOrException) -> ErrCapture:
  err_capture: Optional[ErrCapture] = None

  if issubclass(exception.__class__, ActionError):
    err_capture = exception.err_capture

  if not err_capture:
    err_capture = ErrCapture(
      fatal=True,
      type=exception.__class__.__name__,
      name=exception.__class__.__name__,
      reason=str(exception),
      logs=utils.exception2trace(exception)
    )

  if not err_capture.get('type'):
    err_capture['type'] = 'internal_error'

  if not err_capture.get('name'):
    err_capture['name'] = ''

  if 'is_original' not in err_capture.keys():
    err_capture['is_original'] = True

  err_capture['event_vid'] = vid
  return err_capture


DEBUG_PROPS_KEY = 'debug_props'
TELEM_ENABLED_KEY = 'telem'
TELEM_PROPS_KEY = 'telem_props'
ERROR_TELEM_PROPS_KEY = 'error_telem_props'
ERROR_TELEM_KEY = 'error_telem'

KEY_PARENT_EVENT_VID = 'parent_event_virtual_id'
KEY_PARENT_EVENT_ID = 'parent_event_id'
ACTION_EVENT_TYPE = 'action'
EXPECTED_RUN_ARGS_KEY = 'expects_run_args'



from kama_sdk.model.supplier.predicate.predicate import Predicate


class TruePredicate(Predicate):

  def id(self) -> str:
    return 'sdk.predicate.true'

  def resolve(self) -> bool:
    return True


class FalsePredicate(Predicate):

  def id(self) -> str:
    return 'sdk.predicate.false'

  def resolve(self) -> bool:
    return False



from typing import List, Optional, Dict

import yaml
from typing_extensions import TypedDict

from kama_sdk.core.core.types import K8sResDict, KteaDict, ErrCapture
from kama_sdk.core.ktea.ktea_client import KteaClient
from kama_sdk.core.ktea.ktea_provider import ktea_client
from kama_sdk.model.action.base.action import Action
from kama_sdk.model.action.base.action_errors import FatalActionError
from kama_sdk.model.supplier.ext.biz.resource_selector import ResourceSelector

MaybeResDescs = Optional[List[K8sResDict]]

class OutputFormat(TypedDict):
  res_descs: List[K8sResDict]


class TemplateManifestAction(Action):

  def ktea(self) -> KteaDict:
    return self.resolve_prop('ktea')

  def values(self) -> Dict:
    return self.resolve_prop('values')

  def selectors(self) -> List[ResourceSelector]:
    return self.resolve_prop('selectors')

  def perform(self) -> OutputFormat:
    ktea = self.ktea()
    values = self.values()
    selectors = self.selectors()

    client_inst = ktea_client(ktea=ktea)
    final_values = KteaClient.merge_ktea_vals(ktea, values)

    res_descs = client_inst.template_manifest(final_values)
    scan_failures_and_raise(ktea, res_descs)
    self.add_logs(list(map(yaml.dump, res_descs)))

    filtered_descs = client_inst.filter_res(res_descs, selectors)
    return dict(res_descs=filtered_descs)


def scan_failures_and_raise(ktea: KteaDict, res_descs: MaybeResDescs):
  if res_descs is None:
    ktea_name = ktea.get('uri') if ktea else None
    raise FatalActionError(ErrCapture(
      type='template_manifest_failed',
      reason=f"Templating engine {ktea_name or 'NONE'} returned no data",
      extras=dict(ktea=ktea)
    ))



from typing import Optional, List, TypeVar, Dict, Any

from werkzeug.utils import cached_property

from kama_sdk.core.core import config_man, utils
from kama_sdk.core.core.config_man import config_man, def_vars_key, vndr_inj_vars_key, user_vars_key
from kama_sdk.core.core.utils import dict2keyed
from kama_sdk.model.variable.generic_variable import GenericVariable

T = TypeVar('T', bound='ManifestVariable')


class ManifestVariable(GenericVariable):

  @cached_property
  def mode(self):
    return self.get_prop(MODE_KEY, 'internal')

  @cached_property
  def is_publisher_overridable(self):
    return self.get_prop(RELEASE_OVERRIDE_KEY, False)

  @cached_property
  def tags(self):
    return self.get_prop(TAGS_KEY, [])

  @cached_property
  def default_value(self) -> str:
    defaults = config_man.default_vars()
    return utils.deep_get2(defaults, self.id())

  def is_safe_to_set(self) -> bool:
    return self.mode == 'public'

  def current_value(self, **kwargs) -> Optional[str]:
    manifest_variables = config_man.manifest_variables(**kwargs)
    return utils.deep_get2(manifest_variables, self.id())

  def current_or_default_value(self):
    return self.current_value() or self.default_value

  def is_currently_valid(self) -> bool:
    variables = config_man.manifest_variables()
    is_defined = self.id() in utils.deep2flat(variables).keys()
    crt_val = utils.deep_get2(variables, self.id())
    return self.validate(crt_val)['met'] if is_defined else True

  def sources(self) -> Dict[str, Any]:
    source_types = [user_vars_key, vndr_inj_vars_key, def_vars_key]

    def type2val(key: str) -> Any:
      source: Dict = config_man.read_infer_type(key)
      return utils.deep_get2(source, self.id())

    return {t: type2val(t) for t in source_types}

  @classmethod
  def all_vars(cls) -> List[T]:
    raw = config_man.user_vars()
    committed_vars = dict2keyed(raw)
    models = cls.inflate_all()
    pres = lambda k: len([cv for cv in models if cv.id() == k]) > 0
    for committed_var in committed_vars:
      key = committed_var[0]
      if not pres(key):
        models.append(cls.synthesize_var_model(key))
    return models

  @classmethod
  def publisher_overridable_vars(cls) -> List[T]:
    matcher = lambda cv: cv.is_publisher_overridable
    return list(filter(matcher, ManifestVariable.inflate_all()))

  # noinspection PyBroadException
  @classmethod
  def find_or_synthesize(cls, manifest_variable_id) -> T:
    try:
      return cls.inflate(manifest_variable_id)
    except:
      return cls.synthesize_var_model(manifest_variable_id)

  @staticmethod
  def synthesize_var_model(key: str):
    return ManifestVariable.inflate({
      'id': key,
      MODE_KEY: 'unlisted',
      'title': f'Undocumented Variable {key}',
      'info': f'Undocumented Variable {key}'
    })


RELEASE_OVERRIDE_KEY = 'publisher_overridable'
MODE_KEY = 'mode'
TAGS_KEY = 'tags'



from typing import Optional, Any

from werkzeug.utils import cached_property

from kama_sdk.core.core.types import ConcernAttrMeta
from kama_sdk.core.core.utils import perr
from kama_sdk.model.base.model import Model


class ConcernAttrAdapter(Model):

  def compute(self):
    pass

  @cached_property
  def label(self) -> str:
    return self.get_prop(LABEL_KEY) or self.id()

  def compute_value(self) -> Optional[Any]:
    if VALUE_KEY in self.config.keys():
      return self.resolve_prop(VALUE_KEY, depth=100)
    elif LABEL_KEY in self.config.keys():
      synth_name = f"get::props {CONCERN_KEY}=>{self.label}"
      return self.resolve_prop_value(synth_name, depth=100)
    else:
      perr(self, "given neither label or value!")
      return None

  def as_meta(self) -> ConcernAttrMeta:
    return ConcernAttrMeta(
      title=self.title,
      label=self.label
    )


VALUE_KEY = "value"
LABEL_KEY = "label"
CONCERN_KEY = 'concern'
SET_ID_KEY = 'set_id'



import subprocess
from typing import Dict

from werkzeug.utils import cached_property

from kama_sdk.core.core import utils
from kama_sdk.model.action.base.action import Action


class CmdExecAction(Action):

  CMD_KEY = 'cmd'

  def cmd(self):
    return self.get_prop(self.CMD_KEY, '')

  def perform(self) -> Dict:
    result = subprocess.check_output(self.cmd().split(" "))
    logs = utils.clean_log_lines(result.decode('utf-8'))
    return dict(logs=logs)


def interpolate_token(token: str, buckets) -> str:
  from kama_sdk.core.core.config_man import config_man
  if token == '$app':
    return config_man.ns()
  elif token.startswith("$vars"):
    _, bucket_name, var_name = token.split("::")
    return buckets[bucket_name][var_name]
  else:
    return token



from typing import List, Dict, Any

from kama_sdk.model.action.base.action import Action
from kama_sdk.model.base.model import Model
from kama_sdk.model.operation.operation import Operation


class Concern(Model):

  attr_values: Dict

  def subject(self):
    pass

  def actions(self) -> List[Action]:
    pass

  def operations(self) -> List[Operation]:
    pass


res_constructor_supplier = \
  "sdk.supplier.concern.resource_selector_to_constructors"

ATTR_METAS_KEY = 'attr_metas'
SEED_KEY = 'seed'



from typing import Dict, Optional

from werkzeug.utils import cached_property

from kama_sdk.model.base.model import Model


class Glance(Model):

  @cached_property
  def view_type(self) -> str:
    return self.resolve_prop(VIEW_TYPE_KEY, missing='warn')

  @cached_property
  def legend_icon(self) -> Optional[str]:
    return self.get_prop(LEGEND_ICON_KEY)

  @cached_property
  def legend_type(self) -> str:
    return self.get_prop(LEGEND_TYPE_KEY, 'default')

  @cached_property
  def legend_text(self) -> str:
    return self.get_prop(LEGEND_TEXT_KEY, self.info)

  @cached_property
  def legend_emotion(self) -> Optional[str]:
    return self.get_prop(LEGEND_EMOTION_KEY)

  @cached_property
  def url_intent(self) -> Optional[str]:
    return self.get_prop(URL_INTENT_KEY)

  @cached_property
  def path_intent(self) -> Optional[str]:
    return self.get_prop(PATH_INTENT_KEY)

  def fast_serialize(self) -> Dict:
    return {'id': self.id(), 'title': self.title}

  @cached_property
  def port_forward_spec(self):
    return None

  def gen_legend(self):
    return {
      'text': self.legend_text,
      'type': self.legend_type,
      'icon': self.legend_icon,
      'emotion': self.legend_emotion
    }

  def compute_and_serialize(self):
    return {
      'view_type': self.view_type,
      'title': self.title,
      'legend': self.gen_legend(),
      'spec': self.content_spec(),
      'action': self.url_intent or self.path_intent,
      'port_forward_spec': self.port_forward_spec
    }

  def content_spec(self):
    raise NotImplementedError


master_glances_supplier_id = 'sdk.glance.master_provider'

VIEW_TYPE_KEY = 'view_type'
CONTENT_SPEC_KEY = 'content'
LEGEND_ICON_KEY = 'legend_icon'
LEGEND_TYPE_KEY = 'legend_type'
LEGEND_TEXT_KEY = 'legend_text'
LEGEND_EMOTION_KEY = 'legend_emotion'
URL_INTENT_KEY = 'url'
PATH_INTENT_KEY = 'path'



import validators

from kama_sdk.model.supplier.predicate.predicate import Predicate


class FormatPredicate(Predicate):

  def reason(self) -> str:
    return f"Must be a(n) {self.check_against()}"

  def resolve(self) -> bool:  # should we use unmuck_primitives?
    check = self.check_against()
    challenge = self.challenge()

    if challenge is not None:
      if check in ['integer', 'int', 'number']:
        return type(challenge) == int or challenge.isdigit()
      elif check in ['boolean', 'bool']:
        return str(challenge).lower() not in ['true', 'false']
      elif check == 'email':
        return validators.email(challenge)
      elif check == 'domain':
        return validators.domain(challenge)
    else:
      return False



from datetime import datetime

from kama_sdk.core.core.config_man import config_man
from kama_sdk.core.core.types import ErrCapture
from kama_sdk.core.telem import telem_man
from kama_sdk.model.action.base.action import Action
from kama_sdk.model.action.base.action_errors import ActionError


class CreateBackupAction(Action):
  def perform(self) -> None:
    raise_if_no_db()
    cmap_contents = config_man.serialize()
    telem_man.store_config_backup(dict(
      event_type='backup_action',
      data=cmap_contents,
      timestamp=str(datetime.now())
    ))


# class UpdateLastCheckedAction(Action):
#   def __init__(self, config: Dict):
#     super().__init__(config)
#     self.observer.progress['sub_items'] = [
#       dict(
#         id='update_last_checked',
#         title='Record successful update',
#         info='Record event locally and sync status with Nectar Cloud',
#         status='idle',
#         sub_items=[
#           dict(
#             id='update_config',
#             title='Record successful update timestamp',
#             info='Commit timestamp to master configmap',
#             status='idle'
#           ),
#           dict(
#             id='sync_last_checked',
#             title='Sync status with Nectar Cloud',
#             info='Upload TAM/Wiz metadata',
#             status='idle'
#           )
#         ]
#       )
#     ]
#
#   # noinspection PyBroadException
#   def perform(self):
#     set_sub = lambda *args: self.observer.set_crt_subitem_status(*args)
#     self.observer.set_item_running('update_last_checked')
#
#     set_sub('update_config', 'running')
#     time.sleep(1)
#     config_man.write_last_synced(datetime.now())
#     set_sub('update_config', 'positive')
#
#     set_sub('sync_last_checked', 'running')
#     sync_result = False
#     time.sleep(1)
#     try:
#       telem_man.upload_all_meta()
#       sync_result = True
#     except:
#       print("[kama_sdk::update_last_checked_action] hub rejected sync")
#       print(traceback.format_exc())
#
#     if sync_result:
#       set_sub('sync_last_checked', 'positive')
#       self.observer.set_item_status('update_last_checked', 'positive')
#     else:
#       self.observer.process_error(
#         fatal=False,
#         tone='warning',
#         status='idle',
#         reason='Failed sync status with Nectar Cloud'
#       )

def raise_if_no_db():
  if not telem_man.create_session_if_none():
    raise ActionError(ErrCapture(
      reason='Could not save backup because Telem database not found',
      type='backup_telem_db_offline'
    ))



from typing import List, Union, Any

from k8kat.res.base.kat_res import KatRes
from werkzeug.utils import cached_property

from kama_sdk.core.core import utils
from kama_sdk.core.core.utils import kres2dict
from kama_sdk.model.supplier.base.supplier import Supplier
from kama_sdk.model.supplier.ext.biz.resource_selector import ResourceSelector


class ResourcesSupplier(Supplier):

  @cached_property
  def output_format(self):
    super_value = super().output_format
    if super_value == 'options_format':
      if not self.serializer_type == 'legacy':
        print("[kama_sdk:res_sup] danger using options_format wo legacy ser")
      return dict(id='name', title='name')
    return super_value

  def resource_selector(self) -> ResourceSelector:
    return self.inflate_child(
      ResourceSelector,
      prop=RESOURCE_SELECTOR_KEY
    )

  def _compute(self) -> List[KatRes]:
    result = self.resource_selector().query_cluster()
    return result

  def jq_serialize(self, kr: Union[KatRes, List[KatRes]]) -> Any:
    is_list = utils.listlike(kr)
    new_value = list(map(kres2dict, kr)) if is_list else kres2dict(kr)
    return super().jq_serialize(new_value)


RESOURCE_SELECTOR_KEY = 'selector'



from typing import Dict, List

from werkzeug.utils import cached_property

from kama_sdk.model.base.model import Model


class ConcernSet(Model):

  @cached_property
  def is_one_shot(self) -> bool:
    return self.get_prop(ONE_SHOT_KEY, True)

  def compute_concern_seeds(self):
    return self.get_prop(CONCERN_SEEDS_KEY) or []

  def one_shot_compute(self) -> List[Dict]:
    return []

  @classmethod
  def view_type(cls) -> str:
    return ''


ONE_SHOT_KEY = 'one_shot'
CONCERN_SEEDS_KEY = 'concern_seeds'



from datetime import datetime
from typing import Dict, List, Optional

from kama_sdk.core.core import hub_api_client, utils
from kama_sdk.core.core.config_man import config_man
from kama_sdk.core.core.types import UpdateDict, InjectionsDesc, K8sResDict
from kama_sdk.core.core.utils import deep_merge
from kama_sdk.core.ktea.ktea_provider import ktea_client
from kama_sdk.model.base import pure_provider_ids
from kama_sdk.model.supplier.base.provider import Provider
from kama_sdk.model.supplier.base.supplier import Supplier


def is_using_latest_injection() -> bool:
  bundle = latest_injection_bundle()
  return bundle is None


def fetch_next_update() -> Optional[UpdateDict]:
  config_man.write_last_update_checked(str(datetime.now()))
  return None


def latest_injection_bundle() -> Optional[InjectionsDesc]:
  if config_man.is_real_deployment():
    resp = hub_api_client.get('/injectors/compile')
    if resp.ok:
      # print("INJECTORS ARE")
      # print(resp.json())
      return resp.json()['data']
    else:
      print(f"[kama_sdk::updates_man] err requesting injection {resp.status_code}")
      return None
  else:
    provider_id = pure_provider_ids.mock_injection_bundle_id
    model = Provider.inflate(provider_id)
    return model.resolve() if model else None


def fetch_update(update_id: str) -> Optional[UpdateDict]:
  if config_man.is_real_deployment() and not utils.is_test():
    resp = hub_api_client.get(f'/app_updates/{update_id}')
    if resp.ok:
      return resp.json()['bundle']
    else:
      print(f"[kama_sdk::updates_man] err requesting update {resp.status_code}")
  else:
    model = Provider.inflate(update_id)
    return model.resolve() if model else None


def next_available() -> Optional[UpdateDict]:
  if config_man.is_real_deployment():
    resp = hub_api_client.get(f'/app_updates/available')
    data = resp.json() if resp.status_code < 205 else None
    return data['bundle'] if data else None
  else:
    model = MockUpdate.inflate(mock_update.app_update_id)
    return model.as_bundle()


def fetch_all():
  if config_man.is_real_deployment():
    resp = hub_api_client.get(f'/app_updates')
    return resp.json().get('data') if resp.status_code < 205 else None
  else:
    raise NotImplementedError


def _gen_injection_telem(keys: List[str]):
  all_vars = config_man.user_vars()
  return {k: all_vars[k] for k in keys}


def preview_injection(injection: InjectionsDesc) -> Dict:
  old_defaults = config_man.vnd_inj_vars()
  old_manifest = ktea_client().template_manifest_std()

  kt = Supplier.inflate("sdk.supplier.injections_ktea")

  new_defaults = deep_merge(old_defaults, injection['standard'])

  new_resources = []
  if len(injection['inline']) > 0 and kt:
    new_resources = ktea_client(ktea=kt).dry_run(
      values=injection['inline'],
    )

  new_manifest = [r for r in old_manifest]

  def find_twin(res: K8sResDict) -> Optional[int]:
    for (i, _res) in enumerate(new_manifest):
      if utils.same_res(res, _res):
        return i
    return None

  for new_res in new_resources:
    old_version_ind = find_twin(new_res)
    if old_version_ind:
      old_version = new_manifest[old_version_ind]
      new_manifest[old_version_ind] = deep_merge(old_version, new_res)
    else:
      new_manifest.append(new_res)

  return dict(
    defaults=dict(
      old=old_defaults,
      new=new_defaults
    ),
    manifest=dict(
      old=old_manifest,
      new=new_manifest
    )
  )


def preview(update_dict: UpdateDict) -> Dict:
  old_defaults = config_man.default_vars()
  old_manifest = ktea_client().template_manifest_std()

  new_ktea = updated_release_ktea(update_dict)
  new_ktea_client = ktea_client(ktea=new_ktea)

  new_defaults = new_ktea_client.load_default_values()
  new_manifest_vars = deep_merge(new_defaults, config_man.user_vars())
  new_manifest = new_ktea_client.template_manifest(new_manifest_vars)

  return dict(
    defaults=dict(
      old=old_defaults,
      new=new_defaults
    ),
    manifest=dict(
      old=old_manifest,
      new=new_manifest
    )
  )


def commit_new_ktea(update_dict: UpdateDict):
  new_ktea = updated_release_ktea(update_dict)
  config_man.patch_ktea(new_ktea)


def commit_new_defaults_from_update(update_dict: UpdateDict):
  new_ktea = updated_release_ktea(update_dict)
  new_defaults = ktea_client(ktea=new_ktea).load_default_values()
  config_man.patch_def_vars(new_defaults)


def updated_release_ktea(update: UpdateDict) -> Dict:
  new_ktea = dict(version=update['version'])
  old_ktea = config_man.ktea_desc()

  if new_type := update.get('ktea_type'):
    new_ktea['type'] = new_type
  if new_uri := update.get('ktea_uri'):
    new_ktea['uri'] = new_uri

  if not new_type and not new_type == old_ktea.get('type'):
    msg = f"change from {old_ktea.get('type')} -> {new_type}"
    print(f"[kama_sdk:updates_man] WARN update ktea type {msg}")

  return {**old_ktea, **new_ktea}



from typing import List, Optional

from werkzeug.utils import cached_property

from kama_sdk.model.base.model import Model
from kama_sdk.model.operation.operation_state import OperationState
from kama_sdk.model.operation.step import Step


class Stage(Model):

  STEPS_KEY = 'steps'

  def __init__(self, config):
    super().__init__(config)

  @cached_property
  def description(self):
    return self.get_prop('description')

  @cached_property
  def steps(self) -> List[Step]:
    """
    Loads the Steps associated with the Stage.
    :return: List of Step instances.
    """
    return self.inflate_children(Step, prop=self.STEPS_KEY)

  def step(self, step_id: str) -> Step:
    """
    Finds the Step by key and inflates (instantiates) into a Step instance.
    :param step_id: identifier for desired Step.
    :return: Step instance.
    """
    matcher = lambda step: step.id() == step_id
    return next(filter(matcher, self.steps), None)

  def first_step_id(self) -> Optional[str]:
    """
    Returns the key of the first associated Step, if present.
    :return: Step key or None.
    """
    return self.steps[0].id() if len(self.steps) > 0 else None

  def next_step_id(self, crt_step: Step, op_state: OperationState) -> str:
    """
    Returns the id of the next step, or "done" if no next step exists.
    :param crt_step:
    :param op_state: if-then-else values, if necessary.
    :return: id of next step or "done".
    """
    override_value = crt_step.next_step_id(op_state)
    if override_value:
      return override_value
    else:
      stage_steps = self.steps
      index = step_index(stage_steps, crt_step.id())
      is_not_last = index < len(stage_steps) - 1
      return stage_steps[index + 1].id() if is_not_last else 'done'

def step_index(steps: List[Step], step_id: str) -> int:
  """
  Returns the index of the desired Step.
  :param steps: list of all Steps associated with a Stage.
  :param step_id: id to identify the desired Step.
  :return: index of the desired Step.
  """
  finder = (i for i, step in enumerate(steps) if step.id() == step_id)
  return next(finder)



from typing import List, Dict, Optional, Type

from kama_sdk.core.core.types import K8sResDict
from kama_sdk.core.ktea.ktea_client import KteaClient


class VirtualKteaClient(KteaClient):
  def template_manifest(self, values: Dict) -> List[K8sResDict]:
    return self._template(values)

  def load_default_values(self) -> Dict[str, str]:
    return self._default_values()

  def _template(self, values: Dict) -> List[Dict]:
    raise NotImplementedError

  def _default_values(self) -> Dict:
    return {}


class VirtualKteaClientsMan:
  def __init__(self):
    self._virtual_kteas: List[Type[VirtualKteaClient]] = []

  def add_client(self, virtual_ktea_client: Type[VirtualKteaClient]):
    self._virtual_kteas.append(virtual_ktea_client)

  def find_client(self, **kwargs) -> Optional[Type[VirtualKteaClient]]:
    name = kwargs.pop('name', None)
    if name:
      matcher = lambda vc: vc.__name__ == name
      return next(filter(matcher, self._virtual_kteas), None)
    return None

  def clear(self):
    self._virtual_kteas = []


vktea_clients_man = VirtualKteaClientsMan()



import json
import traceback
from datetime import datetime
from typing import Optional, Dict, Any

from k8kat.res.config_map.kat_map import KatMap
from k8kat.res.secret.kat_secret import KatSecret
from k8kat.utils.main.utils import deep_merge

from kama_sdk.core.core import config_man_helper as helper
from kama_sdk.core.core import utils
from kama_sdk.core.core.types import KteaDict, KamaDict


class ConfigMan:
  def __init__(self):
    self._cmap: Optional[KatMap] = None
    self._ns: Optional[str] = None

  def invalidate_cmap(self):
    self._cmap = None

  def ns(self):
    if not self._ns:
      self._ns = read_ns()
      if not self._ns:
        print("[kama_sdk:config_man:ns] failed to read new NS")
    return self._ns

  def load_master_cmap(self, **kwargs) -> Optional[KatMap]:
    if self.ns():
      reload_strategy = kwargs.pop('reload', None)
      if self.should_reload_cmap(reload_strategy):
        fresh_cmap = KatMap.find(cmap_name, self.ns())
        if fresh_cmap:
          helper.track_cmap_read(self.ns())
        self._cmap = fresh_cmap
      if not self._cmap:
        cmap_id = f"[{self.ns()}/{cmap_name}]"
        print(f"[kama_sdk:config_man:load_cmap] fatal: {cmap_id} nil")
      return self._cmap
    else:
      print("[kama_sdk:config_man:load_cmap] fatal: ns is nil")
      return None

  def _read_entry(self, key: str, **kwargs) -> Any:
    """
    Read in master ConfigMap from Kubernetes API or from
    cache, return entry in data.
    @param key: name of entry in ConfigMap.data
    @return: value at ConfigMap.data[key] or None
    """
    cmap = self.load_master_cmap(**kwargs)
    return cmap.raw.data.get(key) if cmap else None

  def read_infer_type(self, key: str, **kwargs) -> Any:
    raw_value = self._read_entry(key, **kwargs)
    type_mapping = type_mappings.get(key)
    return helper.parse_inbound(raw_value, type_mapping)

  def read_dict(self, key: str, **kwargs) -> Dict:
    raw_val = self._read_entry(key, **kwargs) or '{}'
    return json.loads(raw_val)

  def patch_master_cmap(self, outer_key: str, value: any):
    expected_type = type_mappings.get(outer_key)
    if expected_type:
      if helper.does_type_match(value, expected_type):
        config_map = self.load_master_cmap()
        serialized_value = helper.serialize_outbound(value, expected_type)
        config_map.raw.data[outer_key] = serialized_value
        config_map.touch(save=True)
        helper.track_cmap_write(self.ns())
      else:
        reason = f"[{outer_key}]={value} must be {expected_type} " \
                 f"but was {type(value)}"
        raise RuntimeError(f"[kama_sdk:config_man] {reason}")
    else:
      raise RuntimeError(f"[kama_sdk:config_man] illegal key {outer_key}")

  def app_id(self, **kwargs) -> str:
    return self.read_infer_type(app_id_key, **kwargs)

  def install_id(self, **kwargs) -> str:
    return self.read_infer_type(install_id_key, **kwargs)

  def application_status(self, **kwargs) -> str:
    return self.read_infer_type(status_key, **kwargs)

  def prefs(self, **kwargs) -> Dict:
    return self.read_infer_type(prefs_config_key, **kwargs)

  def ktea_desc(self, **kwargs) -> KteaDict:
    return self.read_infer_type(ktea_config_key, **kwargs)

  def kama_desc(self) -> KamaDict:
    return self.read_infer_type(kama_config_key)

  def default_vars(self, **kwargs) -> Dict:
    return self.read_infer_type(def_vars_key, **kwargs)

  def vnd_inj_vars(self, **kwargs) -> Dict:
    return self.read_infer_type(vndr_inj_vars_key, **kwargs)

  def user_vars(self, **kwargs) -> Dict:
    return self.read_infer_type(user_vars_key, **kwargs)

  def last_updated(self, **kwargs) -> datetime:
    return self.read_infer_type(key_last_updated, **kwargs)

  def manifest_variables(self, **kwargs) -> Dict:
    return utils.deep_merge(
      self.default_vars(**kwargs),
      self.vnd_inj_vars(**kwargs),
      self.user_vars(**kwargs),
    )

  def last_injected(self, **kwargs) -> datetime:
    result = self.read_infer_type(key_last_synced, **kwargs)
    if result:
      return result
    else:
      use_backup = kwargs.pop('or_ancient', True)
      return helper.ancient_dt() if use_backup else None

  def serialize(self):
    config_map = self.load_master_cmap()
    return config_map.raw.data if config_map else {}

  def patch_into_deep_dict(self, dict_key, _vars: Dict, **kwargs):
    crt_dict = self.read_dict(dict_key, **kwargs)
    new_dict = deep_merge(crt_dict, _vars)
    self.patch_master_cmap(dict_key, new_dict)

  def patch_user_vars(self, deep_assignments: Dict[str, any]):
    self.patch_into_deep_dict(user_vars_key, deep_assignments)

  def patch_def_vars(self, assigns: Dict):
    self.patch_into_deep_dict(def_vars_key, assigns)

  def patch_vnd_inj_vars(self, deep_assignments: Dict[str, any]):
    self.patch_into_deep_dict(vndr_inj_vars_key, deep_assignments)

  def patch_prefs(self, assignments: Dict[str, any]):
    self.patch_into_deep_dict(prefs_config_key, assignments)

  def write_last_synced(self, timestamp: datetime):
    self.patch_master_cmap(key_last_updated, timestamp)

  def write_last_injected(self, timestamp: datetime):
    self.patch_master_cmap(key_last_synced, timestamp)

  def write_ktea(self, new_ktea: KteaDict):
    self.patch_master_cmap(ktea_config_key, new_ktea)

  def write_application_status(self, new_status: str):
    self.patch_master_cmap(status_key, new_status)

  def patch_ktea(self, partial_ktea: KteaDict):
    new_ktea = {**self.ktea_desc(), **partial_ktea}
    self.write_ktea(new_ktea)

  def write_manifest_defaults(self, assigns: Dict):
    self.patch_master_cmap(def_vars_key, assigns)

  def is_training_mode(self) -> bool:
    raw_val = self._read_entry(is_training_key)
    return raw_val in ['True', 'true', True]

  def is_real_deployment(self) -> bool:
    if utils.is_test():
      return False
    else:
      return not self.is_training_mode()

  def standard_manifest_values(self) -> Dict:
    return {
      **self.default_vars(),
      **self.user_vars()
    }

  def install_token(self) -> Optional[str]:
    if utils.is_in_cluster():
      try:
        with open(install_token_path, 'r') as file:
          return file.read()
      except FileNotFoundError:
        print(f"[kama_sdk:config_man:install_token] {install_token_path} fnf")
        return None
    else:
      try:
        secret_res = KatSecret.find("master", self.ns())
        return secret_res.decoded_data()['install_token']
      except:
        print(traceback.format_exc())
        print(f"[kama_sdk:config_man:install_token] manual secret fetch ^")
        return None

  def should_reload_cmap(self, strategy: Optional[Any]) -> bool:
    if self._cmap:
      if strategy in [None, 'auto']:
        return helper.is_cmap_dirty(self.ns())
      elif strategy is True:
        return True
      elif strategy is False:
        return False
      else:
        print(f"[kama_sdk:config_man] bad cmap reload strategy {strategy}")
        return True
    else:
      return True


config_man = ConfigMan()


def read_ns() -> Optional[str]:
  """
  Reads application namespace from a file. If in-cluster, path
  will be Kubernetes default
  /var/run/secrets/kubernetes.io/serviceaccount/namespace. Else,
  wiz must be running in dev or training mode and tmp path will be used.
  @return: name of new namespace
  """
  path = ns_path if utils.is_in_cluster() else dev_ns_path
  try:
    with open(path, 'r') as file:
      value = file.read()
      if value:
        value = value.strip()
      else:
        print(f"[kama_sdk::configmap] FATAL ns empty at {path}")
      return value
  except FileNotFoundError:
    print(f"[kama_sdk::configmap] FATAL read failed")
    print(traceback.format_exc())
    return None


def coerce_ns(new_ns: str):
  """
  For out-of-cluster Dev mode only. Changes global ns variable
  to new val, and also writes new val to file so that other processes
  (e.g worker queues) use the same value.
  @param new_ns: name of new namespace
  """
  if utils.is_out_of_cluster():
    config_man._ns = new_ns
    with open(dev_ns_path, 'w') as file:
      file.write(new_ns)
  else:
    print(f"[kama_sdk::configman] illegal ns coerce!")


cmap_name = 'master'

is_training_key = 'is_dev'
app_id_key = 'app_id'
install_id_key = 'install_id'
status_key = 'status'
ktea_config_key = 'ktea'
kama_config_key = 'kama'

def_vars_key = 'default_vars'
vndr_inj_vars_key = 'vendor_injection_vars'
user_inj_vars_key = 'user_injection_vars'
user_vars_key = 'user_vars'
prefs_config_key = 'prefs'

key_last_updated = 'last_updated'
key_last_synced = 'last_synced'
update_checked_at_key = 'update_checked_at'

install_token_path = '/etc/sec/install_token'
mounted_cmap_root_path = '/etc/master_config'
ns_path = '/var/run/secrets/kubernetes.io/serviceaccount/namespace'
dev_ns_path = '/tmp/kama_sdk-dev-ns'
iso8601_time_fmt = '%Y-%m-%d %H:%M:%S.%f'

type_mappings = {
  def_vars_key: dict,
  vndr_inj_vars_key: dict,
  user_vars_key: dict,
  user_inj_vars_key: dict,
  prefs_config_key: dict,
  ktea_config_key: dict,
  kama_config_key: dict,

  key_last_updated: datetime,
  key_last_synced: datetime,
  update_checked_at_key: datetime,

  is_training_key: bool,

  app_id_key: str,
  install_id_key: str,
  status_key: str,
}

manifest_var_keys = [
  user_vars_key,
  vndr_inj_vars_key,
  user_inj_vars_key,
  def_vars_key
]



import functools
import json
import os
import random
import string
import subprocess
import sys
import traceback
import uuid
from datetime import datetime
from functools import reduce
from os import listdir
from os.path import isfile, join, isdir
from pathlib import Path
from typing import Dict, List, Tuple, Optional, Any, Callable

import yaml
from inflection import underscore
from k8kat.auth.kube_broker import broker
from k8kat.res.base.kat_res import KatRes
from k8kat.utils.main.api_defs_man import api_defs_man
from termcolor import cprint
from yaml import scanner

from kama_sdk.core.core.types import KAO, K8sResDict, K8sResSig

legal_envs = ['production', 'development', 'test']


def exec_mode() -> str:
  from_argv = sys.argv[1] if len(sys.argv) >= 2 else None
  return from_argv or 'server'


def exception2trace(exception: Exception) -> List[str]:
  as_list = traceback.format_exception(
    etype=type(exception),
    value=exception,
    tb=exception.__traceback__
  )
  return list(map(str.strip, as_list)) if as_list else []


def is_in_cluster() -> bool:
  return broker.is_in_cluster_auth()


def is_out_of_cluster() -> bool:
  return not is_in_cluster()


def is_shell() -> bool:
  return exec_mode() == 'shell'


def listlike(value: Any) -> bool:
  return isinstance(value, list)


def dictlike(value: Any) -> bool:
  return isinstance(value, dict)


def worker_uuid() -> Optional[str]:
  return os.environ.get('WIZ_WORKER_UUID')


def keyed2dict(keyed_assigns: List[Tuple[str, any]]) -> Dict:
  root = {}
  for keyed_assign in keyed_assigns:
    deep_key_as_list = keyed_assign[0].split('.')  # fully qualified hash key
    deep_set(root, deep_key_as_list, keyed_assign[1])
  return root


def deep_merge(*dicts):
  from k8kat.utils.main.utils import deep_merge as k8kat_deep_merge
  merged = {}
  for _dict in dicts:
    merged = k8kat_deep_merge(merged, _dict)
  return merged


def flat2deep(flat_dict: Dict) -> Dict:
  return keyed2dict(list(flat_dict.items()))


def dict2keyed(assigns: Dict) -> List[Tuple[str, any]]:
  list_keyed_dict = _dict2keyed([], assigns)
  cleaner = lambda t: (".".join(t[0]), t[1])
  return list(map(cleaner, list_keyed_dict))


def deep2flat(assigns) -> Dict:
  tuples = dict2keyed(assigns)
  return {t[0]: t[1] for t in tuples}


def _dict2keyed(parents, assigns: Dict) -> List[Tuple[List[str], any]]:
  result: List[Tuple[List[str], any]] = []

  for assign in assigns.items():
    key, value = assign
    if type(value) == dict:
      result = result + _dict2keyed(parents + [key], value)
    else:
      result.append((parents + [key], value))
  return result


def deep_set(dict_root: Dict, names: List[str], value: any):
  """
  Iterates over items in names list, using them as keys to go deeper into the
  dictionary at each iteration. Eventually sets the passed value with the final key.
  with the passed value.
  :param dict_root: dict to be modified with the desired value.
  :param names: list of names to be iterated over, to find the right depth.
  :param value: value to be eventually set at the right depth.
  """
  if len(names) == 1:
    dict_root[names[0]] = value
  else:
    if not dict_root.get(names[0]):
      dict_root[names[0]] = dict()
    deep_set(dict_root[names[0]], names[1:], value)


def deep_get2(dict_root: Dict, deep_key) -> Any:
  return deep_get(dict_root, deep_key.split('.'))


def deep_set2(dict_root: Dict, deep_key: str, value: Any):
  deep_set(dict_root, deep_key.split("."), value)


def deep_get(dict_root: Dict, keys: List[str]) -> Any:
  """
  Iterates over items in keys list, using them as keys to go deeper into the
  dictionary at each iteration. Eventually retrieves the value of the final key.
  :param dict_root: dict containing the desired value.
  :param keys: list of keys to be iterated over, to find the right depth.
  :return: value of the final key.
  """
  keys = [k for k in keys if not k.strip() == '']

  if len(keys) > 0:
    return reduce(
      lambda d, key: d.get(key, None)
      if isinstance(d, dict)
      else None, keys, dict_root
    )
  else:
    return dict_root


def exclusive_deep_merge(**kwargs) -> Dict:
  old_dict = kwargs['old_dict']
  new_dict = kwargs['new_dict']
  weak_var_keys = kwargs['weak_var_keys']

  merged = deep_merge(old_dict, new_dict)
  old_dict_deep_keys = list(deep2flat(old_dict).keys())
  for deep_key in old_dict_deep_keys:
    if deep_key not in weak_var_keys:
      original_value = deep_get2(old_dict, deep_key)
      deep_set2(merged, deep_key, original_value)

  return merged


def pluck_or_getattr_deep(obj, attr):
  """
  Deep attribute supplier.
  :param obj: Object from which to get the attribute.
  :param attr: attribute to get.
  :return: value of the attribute if found, else None.
  """

  def _getattr(_obj, _attr):
    if type(_obj) == dict:
      return _obj[_attr]
    else:
      if hasattr(_obj, _attr):
        func_or_lit = getattr(_obj, _attr)
        return func_or_lit() if callable(func_or_lit) else func_or_lit
      elif hasattr(_obj, "get_prop"):
        func = getattr(_obj, "get_prop")
        return func(_attr)
  try:
    return functools.reduce(_getattr, [obj] + attr.split('.'))
  except AttributeError:
    return None


def shell_exec(unsplit_cmd: str) -> str:
  formatted_cmd = unsplit_cmd.split(' ')
  output = subprocess.run(
    formatted_cmd,
    stdout=subprocess.PIPE
  )
  return output.stdout.decode('utf-8')


def perr(inst, message, trace=False):
  if trace:
    message = f"{message}\n{traceback.format_exc()}"
  print_colored(inst, message, 'magenta')


def pwar(inst, message, trace=False):
  if trace:
    message = f"{message}\n{traceback.format_exc()}"
  print_colored(inst, message, 'yellow')


def print_colored(inst, message, color):
  if isinstance(inst, type):
    class_sig = underscore(inst.__name__)
  else:
    class_sig = underscore(inst.__class__.__name__)
  cprint(f"[kama_sdk:{class_sig}] {message}", color=color, attrs=['bold'])


def yamls_in_dir(dir_path, recursive=None) -> List[Dict]:
  """
  Parses all YAMLs in the passed dir_path into a single list of dicts.
  :param dir_path: path where YAMLs are located.
  :return: list of dicts.
  """
  is_yaml = lambda path: isfile(path) and path.endswith(".yaml")
  fnames = [f for f in listdir(dir_path) if is_yaml(join(dir_path, f))]
  yaml_arrays = [yamls_in_file(f"{dir_path}/{fname}") for fname in fnames]
  if recursive:
    sub_dirs = [f for f in listdir(dir_path) if isdir(join(dir_path, f))]
    for sub_dir in sub_dirs:
      yaml_arrays.append(yamls_in_dir(join(dir_path, sub_dir), True))
  return flatten(yaml_arrays)


def yamls_in_file(fname) -> List[Dict]:
  """
  Converts a YAML file into a list of dicts.
  :param fname: YAML file.
  :return: list of dicts.
  """
  with open(fname, 'r') as stream:
    file_contents = stream.read()
    try:
      raw_list = list(yaml.full_load_all(file_contents))
      reject_nullish = lambda y: len((y or {}).items()) > 0
      return list(filter(reject_nullish, raw_list))
    except scanner.ScannerError as e:
      print(f"[kama_sdk::utils] YAML parse error @ {fname}")
      raise e


def jk_exec(cmd, **kwargs) -> Dict[str, any]:
  cmd = f"{kmd(cmd, **kwargs)} -o json"
  result = shell_exec(cmd)
  return json.loads(result)


def k_exec(cmd, **kwargs) -> str:
  cmd = kmd(cmd, **kwargs)
  return shell_exec(cmd)


def kmd(cmd: str, **kwargs) -> str:
  kubectl = kwargs.get('k', 'kubectl')
  ns = kwargs.get('ns', None)
  context = kwargs.get('ctx', None)

  cmd = f"{cmd} -n {ns}" if ns else cmd
  cmd = f"{cmd} --context={context}" if context else cmd
  cmd = f"{kubectl} {cmd}"

  return cmd


def root_path() -> str:
  return str(Path(__file__).parent.parent.parent)


def set_run_env(_run_env):
  if _run_env in legal_envs:
    os.environ['FLASK_ENV'] = _run_env
    os.environ['KAT_ENV'] = _run_env
  else:
    raise Exception(f"Bad environment '{_run_env}'")


def run_env() -> str:
  return os.environ.get('FLASK_ENV', 'production')


def is_prod() -> bool:
  return run_env() == 'production'


def is_dev() -> bool:
  return run_env() == 'development'


def is_in_cluster_dev():
  return is_dev() and broker.is_in_cluster_auth()


def is_local_dev_server():
  return not broker.is_in_cluster_auth()


def is_test() -> bool:
  return run_env() == 'test'


def is_ci() -> bool:
  return is_test() and os.environ.get('CI')


def is_ci_keep():
  return os.environ.get("CI") == 'keep'


def any2bool(anything: Any) -> bool:
  if type(anything) == bool:
    return anything
  elif type(anything) == str:
    if anything in ['true', 'True', 'yes', 'positive']:
      return True
    elif anything in ['false', 'False', 'None', 'no', 'negative']:
      return False
    else:
      return bool(anything)
  else:
    return bool(anything)


def is_non_trivial(dict_array):
  if not dict_array:
    return False
  return [e for e in dict_array if e]


def is_either_hash_in_hash(big_hash, little_hashes):
  little_tuples = [list(h.items())[0] for h in little_hashes]
  for _tuple in (big_hash or {}).items():
    if _tuple in little_tuples:
      return True
  return False


def selects_labels(selector: Dict, actual: Dict):
  for key, value in selector.items():
    value_as_list = value if listlike(value) else [value]
    if key not in actual.keys():
      return False
    if actual[key] not in value_as_list:
      return False
  return True


# noinspection PyBroadException
def try_or(lam, fallback=None):
  try:
    return lam()
  except:
    return fallback


def dict_to_eq_str(_dict):
  return ",".join(
    ["=".join([k, str(v)]) for k, v in _dict.items()]
  )


def parse_dict_array(_string):
  parts = _string.split(',')
  return [parse_dict(part) for part in parts]


def parse_dict(encoded_dict):
  result_dict = {}
  for encoded_kv in encoded_dict.split(','):
    key, value = encoded_kv.split(':')
    result_dict[key] = value
  return result_dict


def rand_str(string_len=10):
  letters = string.ascii_lowercase
  return ''.join(random.choice(letters) for _ in range(string_len))


def gen_uuid() -> str:
  return str(uuid.uuid4())


def fqcn(o):
  module = o.__class__.__module__
  if module is None or module == str.__class__.__module__:
    return o.__class__.__name__
  else:
    return module + '.' + o.__class__.__name__


def coerce_cmd_format(cmd):
  if isinstance(cmd, str):
    return cmd.split(" ")
  else:
    return cmd


def clean_log_lines(chunk) -> List[str]:
  if type(chunk) == str:
    log_lines = chunk.split("\n")
    weeder = lambda line: not line.strip() == ''
    return list(filter(weeder, log_lines))
  elif type(chunk) == list:
    return chunk
  else:
    return []


def log2kao(log: str) -> Optional[KAO]:
  try:
    api = ''
    kind_and_name, verb = log.split(" ")
    kind, name = kind_and_name.split("/")
    if "." in kind:
      parts = kind.split(".")
      kind = parts[0]
      api = '.'.join(parts[1:])
    return KAO(
      api_group=api,
      kind=kind,
      name=name,
      verb=verb,
      error=None
    )
  except:
    return None


def kao2log(kao: KAO) -> str:
  group_and_kind = ''
  if kao.get('api_group'):
    group_and_kind = f"{kao.get('api_group')}."
  group_and_kind = f"{group_and_kind}{kao.get('kind')}"
  identity = f"{group_and_kind}/{kao.get('name')}"

  if not kao.get('error'):
    return f"{identity} {kao.get('verb')}"
  else:
    return f"{identity} {kao.get('error')}"


def same_res(r1: K8sResSig, r2: K8sResSig):
  kinds_eq = api_defs_man.kind2plurname(r1['kind']) and \
             api_defs_man.kind2plurname(r2['kind'])
  names_eq = r1['name'] == r2['name']
  return kinds_eq and names_eq


def full_res2sig(res_dict: K8sResDict) -> K8sResSig:
  meta = res_dict.get('metadata') or {}
  return K8sResSig(
    kind=res_dict.get('kind'),
    name=meta.get('name')
  )


def logs2outkomes(logs: List[str]) -> List[KAO]:
  outcomes = list(map(log2kao, logs))
  return [o for o in outcomes if o is not None]


def flatten(nested_list):
  return [item for sublist in nested_list for item in sublist]


def compact(nullish_list: List[Optional[Any]]) -> List[Any]:
  return [item for item in nullish_list if item is not None]


def unmuck_primitives(root: Any) -> Any:
  if type(root) == dict:
    return {k: unmuck_primitives(v) for k, v in root.items()}
  elif type(root) == list:
    return list(map(unmuck_primitives, root))
  else:
    return unmuck_primitive(root)


def isfloat(value):
  try:
    float(value)
    return True
  except ValueError:
    return False


def unmuck_primitive(original: Any) -> Any:
  if type(original) == str:
    if original.isdigit():
      return int(original)
    elif isfloat(original):
      return float(original)
    elif original.lower() == 'true':
      return True
    elif original.lower() == 'false':
      return False
    else:
      return original
  else:
    return original


def kres2dict(kat_res: KatRes) -> Optional[Dict]:
  if kat_res:
    if raw_res := kat_res.raw:
      as_dict = raw_res.to_dict()
      as_dict['kind'] = kat_res.kind
      return sanitize_kres_values(as_dict)
    else:
      print("[kama_sdk:res_sup] danger kat_res has no .raw")
      print(kat_res)
      return sanitize_kres_values(kat_res.__dict__)
  else:
    return None


def safely(func: Callable, bkp=None) -> Any:
  try:
    return func()
  except:
    return bkp


def sanitize_kres_values(root: Any) -> Any:
  if dictlike(root):
    new_root = {}
    for key, value in root.items():
      if listlike(value):
        new_root[key] = list(map(sanitize_kres_values, value))
      elif dictlike(value):
        new_root[key] = sanitize_kres_values(value)
      elif isinstance(value, datetime):
        new_root[key] = str(value)
      else:
        new_root[key] = value
    return new_root
  else:
    return root
